//import to use functionalities
using UnityEngine;
using System.IO;
using Dummiesman;
using System;

public class MeshImporter3 : MonoBehaviour
{
    void Awake()
    {
    	//Path for the log file, in order to see if everything worked as expected. If e.g. no mesh can be seen in a GUI test, then check the log, if the path to the mesh was found and maybe the positioning is off (not in the view of the camera)...
        string logFilePath = "/home/akreibich/TestRobinCode37/DebugLogs.txt";
        File.AppendAllText(logFilePath, "Awake started.\n");

        try
        {
            //Here we initialize the mesh path variables, for this we need a way to parse the name from the command line environment that gets simulated by the python script.
            string[] args = Environment.GetCommandLineArgs();
            File.AppendAllText(logFilePath, "Received arguments: " + string.Join(", ", args) + "\n");

            string meshPath ="";
            string secondMeshPath =  "";
		//If one runs via the python loop, it should be like this: (so here we initialize transformation vectors to zero), if the objective is to test it in the scene itself, set the position manually below by commenting it out and placing it correctly).
		Vector3 position = Vector3.zero;
		Vector3 scale = Vector3.one;
		Vector3 rotation = Vector3.zero;

		//Change it to this for testing
		//Vector3 position = new Vector3(1311f, -392f, -1197f);
		//Vector3 scale = new Vector3(27.2f, 27.2f, 27.2f);
		//Vector3 rotation = new Vector3(0f, -86.6f, 0f);

            //Iterate through command line arguments to set the above initialized mesh paths and transformations for the positioning of the mesh.
            //Adding a second .obj file is possible, as a second mesh if the goal is to not allow collisions from the outside and from the inside. Then the second mesh is the same mesh as the first but with its normals inverted.
            for (int i = 0; i < args.Length; i++)
            {
                if (args[i] == "-objPath")//This checks if the argument is for object path, so for the position, scale, paths... and everytime it finds the name, it takes the next values as the argument.
                {
                    meshPath = args[i + 1];
                }
                else if (args[i] == "-secondObjPath")
                {
                    secondMeshPath = args[i + 1];
                }
                else if (args[i] == "-position")
                {
                    //Split the argument into components and create a vector
                    string[] posArgs = args[i + 1].Split(',');
                    position = new Vector3(float.Parse(posArgs[0]), float.Parse(posArgs[1]), float.Parse(posArgs[2]));
                }
                else if (args[i] == "-scale")
                {
                    //Split the argument into components and create a vector
                    string[] scaleArgs = args[i + 1].Split(',');
                    scale = new Vector3(float.Parse(scaleArgs[0]), float.Parse(scaleArgs[1]), float.Parse(scaleArgs[2]));
                }
                else if (args[i] == "-rotation")
                {
                    //Again we split the argument into components and create a vector
                    string[] rotArgs = args[i + 1].Split(',');
                    rotation = new Vector3(float.Parse(rotArgs[0]), float.Parse(rotArgs[1]), float.Parse(rotArgs[2]));
                }
            }

            if (string.IsNullOrEmpty(meshPath))
            {
                Debug.LogError("Primary mesh path not specified");
                File.AppendAllText(logFilePath, "Primary mesh path not specified\n");
                return;
            }

            Mesh mesh = LoadMesh(meshPath, logFilePath);
            Mesh secondMesh = string.IsNullOrEmpty(secondMeshPath) ? null : LoadMesh(secondMeshPath, logFilePath);

            ImportMesh(mesh, secondMesh, position, scale, rotation);
        }
        catch (Exception e)
        {
            Debug.LogError("An exception occurred: " + e.ToString());
            File.AppendAllText(logFilePath, "An exception occurred: " + e.ToString() + "\n");
        }
    }

    Mesh LoadMesh(string meshPath, string logFilePath)
    {
        OBJLoader objLoader = new OBJLoader();
        GameObject loadedObj = objLoader.Load(meshPath);

        if (loadedObj == null)
        {
            Debug.LogError("Failed to load mesh from path: " + meshPath);
            File.AppendAllText(logFilePath, "Failed to load mesh from path: " + meshPath + "\n");
            return null;
        }

        File.AppendAllText(logFilePath, "Successfully loaded mesh from path: " + meshPath + "\n");

        MeshFilter meshFilter = loadedObj.GetComponentInChildren<MeshFilter>();
        Destroy(loadedObj);

        if (meshFilter == null)
        {
            Debug.LogError("No MeshFilter component found in loaded GameObject");
            File.AppendAllText(logFilePath, "No MeshFilter component found in loaded GameObject\n");
            return null;
        }

        return meshFilter.sharedMesh;
    }

    void ImportMesh(Mesh primaryMesh, Mesh secondaryMesh, Vector3 position, Vector3 scale, Vector3 rotation)
    {
        if (primaryMesh == null)
        {
            Debug.LogError("Primary mesh is null");
            return;
        }

        //This creates a new GameObject and sets its transformation properties: (these are defined either above or parsed from the outside loop)
        GameObject obj = new GameObject(primaryMesh.name);
        obj.transform.position = position;
        obj.transform.localScale = scale;
        obj.transform.eulerAngles = rotation;
        

        //Add the MeshFilter component and assign the primary mesh, do the same below for the MeshRenderer component below. Then the material is loaded, in order for the artery to have the typical optic and seem realistic
        MeshFilter filter = obj.AddComponent<MeshFilter>();
        filter.mesh = primaryMesh;

        MeshRenderer renderer = obj.AddComponent<MeshRenderer>();
        Material mat = Resources.Load<Material>("Aorta Material");

        if (mat != null)
        {
            renderer.material = mat;
        }
        else
        {
            Debug.LogError("Failed to load material");
        }

        //Now add a MeshCollider and the primary mesh, in order to enable collision detection between guidewire and artery that can be resolved in the constraint solving.
        MeshCollider primaryCollider = obj.AddComponent<MeshCollider>();
        primaryCollider.sharedMesh = primaryMesh;

        if (secondaryMesh != null)
        {
            MeshCollider secondaryCollider = obj.AddComponent<MeshCollider>();
            secondaryCollider.sharedMesh = secondaryMesh;
        }

        //Rigidbody component needs to be added as well in order to enable the collision detection, use isKinematic on, for all colliding objects, then the collision data will be output and not corrected by the Unity engine itself
        Rigidbody rb = obj.AddComponent<Rigidbody>();
        rb.useGravity = false;
        rb.isKinematic = true;

        //adding the correct layer for the layer matrix to have the correct effect
        int layerIndex = LayerMask.NameToLayer("Blood Vessel");

        if (layerIndex == -1)
        {
            Debug.LogError("Layer 'Blood Vessel' not found");
        }
        else
        {
            obj.layer = layerIndex;
        }
    }
}

